/* ASELoader.c
 * $Format: " * Version $ProjectVersion$"$
 * Version 0.27
 * Thomas Morrissey, tm7733@bris.ac.uk
 * 15/11/99
 * Routines to read and parse an ASE file and generate the object information
 * 
 * Comments
 * --------
 *
 * Not all fields are fully supported at the moment, however they will be added later as needed, namely:
 *
 *  Sub materials (present in some complex meshes exported from 3D Studio Max)
 *
 * Potential Bugs/problems with parser
 * -----------------------------------
 *
 * Will generate error for 0 vertexes -> 0 Vertices is unlikely/impossible
 *
 */ 

#ifdef WIN32
  #include <windows.h>
  #include <sys/timeb.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "Constants.h"
#include "Lexical.h"
#include "ASELoader.h"
#include "ASETokens.h"
#include "Object.h"
#include "Camera.h"
#include "Materials.h"

#ifdef CN_Error
 #undef CN_Error
#endif
#define CN_Error(a)  Error(a,__FILE__,__LINE__,File)

#define Expect(type)   if(File->NextToken!=type) \
                         CN_Error("Unexpected token"); \
                       GetNextToken(File); 

#define GROUP_CHUNKSIZE   10    /* Default size of array for groups */

/* ASE File type */
struct asefile
{
  char         FileName[256];   /* File name of the ASE file        */
  char         Path[50];        /* Path for the file                */
  int         *MatIDReference;  /* The material reference ID        */
  long         FileSize;        /* Size of the ase file             */
  lexreader_t  reader;          /* Reader for the lexical analyser  */
  Token_t      NextToken;       /* The current token being examined */
  FILE        *Depend;          /* Stream for writing the dependencies */
};

/* Internal function prototypes */
static boolean    ExportMode        (struct asefile *File);
static boolean    Comment           (struct asefile *File);
static boolean    Scene             (struct asefile *File);
static boolean    EnvMap            (struct asefile *File);
static boolean    MaterialList      (struct asefile *File);
static boolean    Material          (struct asefile *File, int MaterialID);
static object_t  *GeomObject        (struct asefile *File);
static camera_t  *CameraObject      (struct asefile *File);
static shape_t   *TrackObject       (struct asefile *File);
static asegroup_t*GroupObject       (struct asefile *File);
static void       MeshVertexList    (struct asefile *File, int VertexID, object_t *NewObject);
static void       MeshFaceList      (struct asefile *File, int FaceID,   object_t *NewObject);
static void       MeshTVert         (struct asefile *File, int TVertID,  object_t *NewObject);
static void       MeshTFaceList     (struct asefile *File, int TFaceID,  object_t *NewObject);
static void       MeshNormals       (struct asefile *File, int NormalID, object_t *NewObject);
static void       TextureVertex     (struct asefile *File, object_t *NewObject);
static void       ColourVertex      (struct asefile *File, object_t *NewObject);
static void       GetNextToken      (struct asefile *File);
static void       NodeTM            (struct asefile *File,vec3_t *Pos);
static void       Error             (const char *Message,const char *File,int Line, struct asefile *asefile);
static void       Depend            (struct asefile *File, char *message);


/*GetNextToken*/
/* Reads the next token into the global variable */
void GetNextToken(struct asefile *File)
{
  int i;
  
  /* Get the next lecial token */
  File->NextToken = LEX_ReadToken(&(File->reader));
  
  /* If token is an identifier, then find out what type */
  if(File->NextToken == IDENTIFIER)
  {
    /* Traverse the list of symbols looking for a match
     */
    for(i = 0; ASE_SymbolNameList[i] != NULL; i++)
      if(strcmp(ASE_SymbolNameList[i], File->reader.TokenValue.s) == 0)
      {
        File->NextToken = i;       /* This corresponds to its symbol */
        return;
      }
      File->NextToken = BAD;     /* If no match found then must be BAD */
  }
}

/*ASE_Open*/
/* Opens file, and loads into memory 
 */
boolean ASE_Open(const char *Path,const char *Filename, asefile_t *New)
{
  char  charbuf[256];
  char  namebuf[50];
  FILE *Stream;          /* Stream containing the file data */

  /* Create memory for the new object */
  (*New) = calloc(sizeof(struct asefile),1);

  /* Concat the path with the filename */
  sprintf(charbuf, "%s%s", Path, Filename);

  /* Open a stream for the file in RTEXT mode*/
  if((Stream=fopen(charbuf, "r")) == NULL)
  {
    printf("Could not open ASE file: %s%s\n", Path, Filename);
    return false;
  }

  /* Initialise its values */
  strcpy((*New)->FileName, Filename);
  strcpy((*New)->Path, Path);

  /* Create a lexical reader from an io stream */
  if(!LEX_InitReaderIO(&((*New)->reader), Stream))
  {
    printf("Could not initialise lexical reader for: %s\n", Filename);
    return false;
  }

  /* Now load the data into memory */
  printf("Loading %s\n", Filename);

  /* Open the stream for writing the dependencies */
  /* It it fails, then simply ignore, since directory may be read only, i.e, CDROM */
  printf("Filename: %s\n",Filename);
  if(strlen(Filename) > 5)
  {
    time_t ltime;

    strncpy(namebuf, Filename, strlen(Filename) - 4);
    namebuf[strlen(Filename) - 4] = '\0';
    sprintf(charbuf, "%s%s_depend.txt", Path, namebuf);
    (*New)->Depend = fopen(charbuf, "w");

    time( &ltime );
    sprintf(charbuf, "File automatically generated by Rogue Scavangers, version %s, %s","22", ctime(&ltime));
    Depend(*New, charbuf);
    sprintf(charbuf, "Dependencies for %s\n\n", (*New)->FileName);
    Depend(*New, charbuf);
  }

  GetNextToken(*New); 
  ExportMode(*New);
  
  while((*New)->NextToken == COMMENT)
    Comment(*New);
  
  if(Scene(*New) == false)      /* Error encountered within SCENE */
    return false;

  switch((*New)->NextToken)
  {
    case MATERIAL_LIST:
      if(MaterialList(*New) == false)   /* If an error encountered in MaterialList non-terminal */
        return false;
      else
        return true;
    default:
      CN_RealWarning("MaterialList, GEOMOBJECT or LIGHTOBJECT expected",__FILE__,__LINE__);
      return false;
  }
}

/*ASE_GetObject*/
/* Reads and returns the next object in the file
 */
aseobject_t ASE_GetObject(struct asefile *File)
{
  aseobject_t NewASEObj;

  /* Read in each object and pass to the Quadtree module */
  switch(File->NextToken)
  {
  /* Geomobject containing a collection of polygon faces */
  case GEOMOBJECT:
    GetNextToken(File);

    NewASEObj.Type=ASE_GeomObject;
    NewASEObj.d.GeomObject=GeomObject(File);
    return NewASEObj;
 
  /* Camera views used for overlooking the world */
  case CAMERAOBJECT:
    GetNextToken(File);

    NewASEObj.Type=ASE_Camera;
    NewASEObj.d.Camera=CameraObject(File);
    return NewASEObj;
 
  /* Line paths used for drone ships */
  case SHAPEOBJECT:
    GetNextToken(File);

    NewASEObj.Type = ASE_Shape;
    NewASEObj.d.Track = TrackObject(File);
    return NewASEObj;
  
  /* Not currently supported so ignore for the moment */
  case LIGHTOBJECT:
    NewASEObj.Type = ASE_Light;
    GetNextToken(File);
    while(File->NextToken != GEOMOBJECT   && File->NextToken != CAMERAOBJECT &&
          File->NextToken != GROUPOBJECT  && File->NextToken != SHAPEOBJECT  &&
          File->NextToken != HELPEROBJECT && File->NextToken != LIGHTOBJECT)
          GetNextToken(File);
    return NewASEObj;

  /* Helper object - don't really know what this does, and not needed anyway so ignore */
  case HELPEROBJECT:
    NewASEObj.Type = ASE_Helper;
    GetNextToken(File);
    while(File->NextToken != GEOMOBJECT   && File->NextToken != CAMERAOBJECT &&
          File->NextToken != GROUPOBJECT  && File->NextToken != SHAPEOBJECT  &&
          File->NextToken != HELPEROBJECT && File->NextToken != LIGHTOBJECT)
          GetNextToken(File);
    return NewASEObj;

  /* Group containing further objects */
  case GROUPOBJECT:
    GetNextToken(File);
    NewASEObj.Type = ASE_Group;
    NewASEObj.d.Group = GroupObject(File);
    return NewASEObj;

  /* End of file token */
  case EOF:
    NewASEObj.Type = ASE_EOF;
    return NewASEObj;

  default:
    printf("Unrecognised object in ASE file\n");
    NewASEObj.Type = ASE_Invalid;
    return NewASEObj;
  }

  return NewASEObj;
}

/*ASE_Close*/
/* Close and free the ADT object
 */
void ASE_Close(struct asefile *File)
{
  if(File->reader.DataType == IO)
    fclose(File->reader.data.io.Stream);
  free(File->MatIDReference);
  if(File->Depend != NULL)
    fclose(File->Depend);
  free(File);
}

/*ExportMode*/
static boolean ExportMode(struct asefile *File)
{
  /* It is assumed that this string is unqiue to ASE files exported from 3DSMAX
   * Therefore it is not an essential requirement and its absence should not imply
   * a syntax error
   */
  if(File->NextToken == TDSMAX_ASCIIEXPORT)
  {
    GetNextToken(File);
    if(File->NextToken!=INTEGER)
    {
      CN_Warning("Version number missing\n");
      return false;
    }
    else
    {
     GetNextToken(File);
    }
  }
  return true;            /* Not false, because absence is not an unrecoverable error */
}

/*Comment*/
/* Reads comments from the file */
static boolean Comment(struct asefile *File)
{  
  GetNextToken(File);
  if(File->NextToken!=STRING)
  {
    CN_Warning("Comment missing\n");
    return false;
  }
  else
  {
    GetNextToken(File);
    return true;
  }
}

/*Scene*/
/* Reads the *SCENE data */
static boolean Scene(struct asefile *File)
{
  if(File->NextToken != SCENE)
  {
    CN_Warning("No SCENE found in file");
    return true;                          /* Does no insinuate an error, SCENE may be optional */
  }

  GetNextToken(File);
  if(File->NextToken!=BRACE_OPEN)
  {
    CN_Warning("Opening Brace expected");
    return false;
  }

  /* Skip most the data at the moment, if wanted additional routines must be written */
  GetNextToken(File);
  while(File->NextToken!=BRACE_CLOSE)
  {
    GetNextToken(File);
    if(File->NextToken == SCENE_ENVMAP)
    {
      GetNextToken(File);
      if(EnvMap(File) == false)
        return false;
    }

  }
  GetNextToken(File);
  return true;
}


/*EnvMap*/
/* Reads the environment map file, like a background image
 */
static boolean EnvMap(struct asefile *File)
{
  material_t SkyMaterial;

  Expect(BRACE_OPEN);

  while(File->NextToken!=BITMAPs)
    GetNextToken(File);

  GetNextToken(File);
  SkyMaterial.Filename = strdup(CN_ExtractFileName(File->reader.TokenValue.s));

  GetNextToken(File);
  while(File->NextToken!=UVW_U_OFFSET)
    GetNextToken(File);

  GetNextToken(File);
  SkyMaterial.uOffset = File->reader.TokenValue.f;

  GetNextToken(File);
  Expect(UVW_V_OFFSET)
  
  SkyMaterial.vOffset = File->reader.TokenValue.f;

  GetNextToken(File);
  Expect(UVW_U_TILING)

  SkyMaterial.uTile = File->reader.TokenValue.f;

  GetNextToken(File);
  Expect(UVW_V_TILING)

  SkyMaterial.vTile = File->reader.TokenValue.f;

 // HACK!!
  /* I have retained tex co-ords for now as the ones in the ASE file are
   *   not in polar form as needed for our sky dome...
   */
  SkyMaterial.uTile         = 1.0;
  SkyMaterial.vTile         = (float) 0.3;
  SkyMaterial.uOffset       = 0.0;
  SkyMaterial.vOffset       = 0.0;
  SkyMaterial.Diffuse[0]    = (float) .7;
  SkyMaterial.Diffuse[1]    = (float) .7;
  SkyMaterial.Diffuse[2]    = (float) 1.0;
  SkyMaterial.Transparency  = 1.0;

  SkyMaterial.Path = strdup(File->Path);

  while(File->NextToken!=BRACE_CLOSE)
    GetNextToken(File);

  /* Ready token for next section */
  GetNextToken(File);

  /* Now Call SkySetup function */
  //SKY_Setup(SkyMaterial);

  return true;
}

/*MaterialList*/
/* Reads the list of materials */
static boolean MaterialList (struct asefile *File)
{
  int           MaterialCount;
  int           i;
  char          charbuf[128];
  
  if(File->NextToken!= MATERIAL_LIST)
    CN_Warning("Material list expected");
  GetNextToken(File); 

  if(File->NextToken!=BRACE_OPEN)
  {
    CN_Warning("Opening Brace expected");
    return false;
  }
  GetNextToken(File);

  if(File->NextToken!=MATERIAL_COUNT)
  {
    CN_Warning("Material count expected");
    return false;
  }
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
  {
    CN_Warning("Integer expected");
    return false;
  }

  MaterialCount = File->reader.TokenValue.i;
  GetNextToken(File);

  sprintf(charbuf, "Number of materials: %d\nTextures required:-\n\n", MaterialCount);
  Depend(File, charbuf);

  /* Read the materials */
  if(MaterialCount>0) {
    if((File->MatIDReference = malloc(sizeof(int) * MaterialCount)) == NULL)
    {
      CN_Warning("Memory allocation for material failed");
      return false;
    }

    for(i = 0; i < MaterialCount; i++)
      if(Material(File,i) == false)
        return false;
  }

  if(File->NextToken!=BRACE_CLOSE)
  {
    CN_Warning("Brace expected");
    return false;
  }
  
  GetNextToken(File);
  return true;
}

/*Material*/
/* Reads in the details for a particular material */
static boolean Material(struct asefile *File, int MaterialID)
{
  char          Message[256];
  material_t    NewMaterial;
  boolean       TextureFile;


  if(File->NextToken!=MATERIAL)
  {
    CN_Warning("Material expected");
    return false;
  }
  GetNextToken(File); 

  if(File->NextToken!=INTEGER)
  {
    CN_Warning("Integer for Material ID expected");
    return false;
  }
  GetNextToken(File); 

  if(MaterialID != File->reader.TokenValue.i)
  {
    sprintf(Message, "Inconsistent material ID's, got %d expected %d",File->reader.TokenValue.i, MaterialID);
    CN_Warning(Message);
    return false;
  }

  if(File->NextToken!=BRACE_OPEN)
  {
    CN_Warning("Brace expected");
    return false;
  }
  
  GetNextToken(File); 

  if(File->NextToken!=MATERIAL_NAME)
  {
    CN_Warning("Material Name expected");
    return false;
  }
  GetNextToken(File); 

  if(File->NextToken!=STRING)
  {
    CN_Warning("String, material name expected");
    return false;
  }
  NewMaterial.Name = strdup(File->reader.TokenValue.s);
  
  GetNextToken(File); 
  
  if(File->NextToken!=MATERIAL_CLASS)
  {
    CN_Warning("Material Class expected");
    return false;
  }
  GetNextToken(File); 

  NewMaterial.Class = strdup(File->reader.TokenValue.s);
  GetNextToken(File); 

  if(File->NextToken!=MATERIAL_AMBIENT)
  {
    CN_Warning("ambient Class expected");
    return false;
  }
  GetNextToken(File); 

  NewMaterial.Ambient[0] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Ambient[1] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Ambient[2] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Ambient[3] = 1.0;

  if(File->NextToken!=MATERIAL_DIFFUSE)
  {
    CN_Warning("diffuse Class expected");
    return false;
  }
  GetNextToken(File); 
  NewMaterial.Diffuse[0] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Diffuse[1] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Diffuse[2] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Diffuse[3] = 1.0;

  if(File->NextToken!=MATERIAL_SPECULAR)
  {
    CN_Warning("specular Class expected");
    return false;
  }
  GetNextToken(File); 
  NewMaterial.Specular[0] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Specular[1] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Specular[2] = File->reader.TokenValue.f;
  GetNextToken(File); 
  NewMaterial.Specular[3] = 1.0;

  if(File->NextToken!=MATERIAL_SHINE)
  {
    CN_Warning("shine value expected");
    return false;
  }
  GetNextToken(File); 
  NewMaterial.Shine = File->reader.TokenValue.f;
  GetNextToken(File); 

  Expect(MATERIAL_SHINESTRENGTH);
  Expect(FLOAT);

  if(File->NextToken!=MATERIAL_TRANSPARENCY)
  {
    CN_Warning("transparency value expected");
    return false;
  }
  GetNextToken(File); 
  NewMaterial.Transparency=1-File->reader.TokenValue.f;
  GetNextToken(File); 

  /* Most of the successive fields can be skipped
   * The only ones of concern are the texture, filename
   * and UV, (texture co-ordinates. However, these are
   * not always present in a material
   */

  while(File->NextToken!=MATERIAL_XP_TYPE)
    GetNextToken(File);
  GetNextToken(File);

  /* Check for presence of MAP_DIFFUSE section */
  GetNextToken(File);

  if(File->NextToken == MAP_DIFFUSE)
  {
    /* Should be in a new function now.... */

    /* Assume there is a texture file bound to the material */
    TextureFile = true;

    /* Now search for the presence of a BITMAP field before the next closing brace */
    while(File->NextToken!=BITMAPs)
    {
      if(File->NextToken == BRACE_CLOSE)
      {
        TextureFile = false;
        break;
      }
      GetNextToken(File);
    }

  } else {
    /* We have probably just read the last }, so we should exit */
    NewMaterial.Filename=NULL;
    NewMaterial.Path = NULL;
    File->MatIDReference[MaterialID] = MAT_Store(NewMaterial);

    GetNextToken(File);
    return true;
  }

  /* End of hack section. mcm */

    /* There is a texture to be bound to the material, so get the filename and UV co-ordinates */
  if(TextureFile == true)
  {
    char charbuf[50];

    GetNextToken(File);
    if(File->NextToken!=STRING)
    {
      CN_Warning("String for filename expected");
      return false;
    }
    NewMaterial.Filename = strdup(CN_ExtractFileName(File->reader.TokenValue.s));
    NewMaterial.Path     = strdup(File->Path);
    sprintf(charbuf, "%s\n", NewMaterial.Filename);
    Depend(File, charbuf);    

    while(File->NextToken!=UVW_U_OFFSET)
      GetNextToken(File);
    GetNextToken(File);
    if(File->NextToken!=FLOAT)
    {
      CN_Warning("Float expected ");
      return false;
    }
    NewMaterial.uOffset = File->reader.TokenValue.f; 
  
    GetNextToken(File);
    if(File->NextToken!=UVW_V_OFFSET)
    {
      CN_Warning("UVW_V_OFFSET expected");
      return false;
    }
    GetNextToken(File);
    if(File->NextToken!=FLOAT)
    {
      CN_Warning("Float expected ");
      return false;
    }
    NewMaterial.vOffset = File->reader.TokenValue.f;

    GetNextToken(File);
    if(File->NextToken!=UVW_U_TILING)
    {
      CN_Warning("UVW_U_TILING expected");
      return false;
    }
    GetNextToken(File);
    if(File->NextToken!=FLOAT)
    {
      CN_Warning("Float expected");
      return false;
    }

    NewMaterial.uTile = File->reader.TokenValue.f;

    GetNextToken(File);
    if(File->NextToken!=UVW_V_TILING)
    {
      CN_Warning("UVW_V_TILING set expected");
      return false;
    }
    GetNextToken(File);
    if(File->NextToken!=FLOAT)
    {
      CN_Warning("Float expected ");
      return false;
    }
    NewMaterial.vTile = File->reader.TokenValue.f;

    while(File->NextToken!=BRACE_CLOSE) 
      GetNextToken(File); 
  } else {
    /* No BITMAP field was found, so obviously no texture to bind */
    NewMaterial.Filename = NULL;
  }

  /* Add the material to the material list */
  File->MatIDReference[MaterialID] = MAT_Store(NewMaterial);

  GetNextToken(File); 
  
  /* The closing brace for the material section */
  while(File->NextToken != BRACE_CLOSE)
    GetNextToken(File); 
 
  /* As per convention, leave current token as start of next expression */
  GetNextToken(File);
  return true;
}

/*GeomObject*/
/* Reads in the geometric objects and creates vertexes */
static object_t *GeomObject(struct asefile *File)
{
  object_t     *NewObject;
  int           i;
  char          charBuf[100];

  if((NewObject = calloc(sizeof(object_t),1)) == NULL)
    CN_Error("Memory allocation failed");

  if(File->NextToken!=BRACE_OPEN)
    CN_Error("opening brace expected");
  GetNextToken(File);

  if(File->NextToken!=NODE_NAME)
    CN_Error("Node name expected");
  GetNextToken(File);

  if(File->NextToken!=STRING)
    CN_Error("name expected");
  NewObject -> Name = strdup(File->reader.TokenValue.s);
  GetNextToken(File);

  /* Skip the TM section */
  while(File->NextToken!=BRACE_CLOSE)
    GetNextToken(File);
  GetNextToken(File);

  if(File->NextToken!=MESH)
    CN_Error("mesh name expected");
  GetNextToken(File);

  if(File->NextToken!=BRACE_OPEN)
    CN_Error("opening brace exected");
  GetNextToken(File);

  if(File->NextToken!=TIMEVALUE)
    CN_Error("timevalue  exected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesth time value exected");
  GetNextToken(File);

  if(File->NextToken!=MESH_NUMVERTEX)
    CN_Error("Mesh vertex number exected");
  GetNextToken(File);
 
  if(File->NextToken!=INTEGER)
    CN_Error("mesh vertex number expected");
  NewObject -> NumVertex = File->reader.TokenValue.i;
  GetNextToken(File);

  if(File->NextToken!=MESH_NUMFACES)
    CN_Error("mesh face number expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");
  NewObject -> NumFaces = File->reader.TokenValue.i;
  GetNextToken(File);

  GetNextToken(File); /* MESH_VERTEX_LIST */
  GetNextToken(File); /* BRACE_OPEN */
  
  /* may want to check for 0 vertexes, unlikely, but will break parser */
  /* Read the vertex list */
  if((NewObject -> Verts = malloc(sizeof(vec3_t) * NewObject -> NumVertex)) == NULL)
    CN_Error("Memory allocation failed");
  for(i = 0; i < NewObject -> NumVertex; i++)
    MeshVertexList(File, i, NewObject);
  
  GetNextToken(File); /* BRACE_CLOSE */
  GetNextToken(File); /* MESH_FACE_LIST */
  GetNextToken(File); /* BRACE_OPEN */
  if((NewObject -> Faces = malloc(sizeof(triangle_t) * NewObject -> NumFaces)) == NULL)
    CN_Error("Memory allocation failed");
  for(i = 0; i < NewObject -> NumFaces; i++)
    MeshFaceList(File, i, NewObject);
  GetNextToken(File); /* BRACE_CLOSE */

  /* Texture vertices */
  if(File->NextToken == MESH_NUMTVERTEX)
  {
    GetNextToken(File);
    TextureVertex(File,NewObject);
  }
  else
  {
    NewObject->NumTVertex=0;
    printf(charBuf, "Warning: Texture co-ordinates missing for "
                         "object \"%s\"\n",NewObject->Name);
  }

  /* Coloured vertices */
  if(File->NextToken == MESH_NUMCVERTEX) {
    GetNextToken(File);
    ColourVertex(File,NewObject);  
  }

  /* Vertex normals */
  if((NewObject -> VertexNorms = malloc(sizeof(vec3_t) * NewObject -> NumVertex)) == NULL)
    CN_Error("Memory allocation failed\n");

  if(File->NextToken!=MESH_NORMALS)
    CN_Error("mesh normals expected");

  GetNextToken(File); 
  if(File->NextToken!=BRACE_OPEN)
    CN_Error("Opening brace was expected");

  GetNextToken(File);
  
  for(i = 0; i < NewObject -> NumFaces; i++)
    MeshNormals(File, i, NewObject);

  if(File->NextToken!=BRACE_CLOSE)
    CN_Error("A closing brace was missing\n");
  
  GetNextToken(File);
  if(File->NextToken!=BRACE_CLOSE)
    CN_Error("A closing brace was missing\n");

  do {
    GetNextToken(File);
  } while(File->NextToken!=MATERIAL_REF && File->NextToken!=BRACE_CLOSE);

  /* HACK! This should be in a new function */  
  if(File->NextToken == MATERIAL_REF) {

    GetNextToken(File);
    if(File->NextToken !=INTEGER)
      CN_Error("Integer expected");
    NewObject->MId = File->MatIDReference[File->reader.TokenValue.i];

    GetNextToken(File);
    if(File->NextToken!=BRACE_CLOSE)
      CN_Error("Closing brace expected");
  
  } else {
    NewObject->MId=-1;
    printf(charBuf, "Warning: No material for object \"%s\"",NewObject->Name);
  }

  GetNextToken(File);

  return(NewObject);
}


/*TextureVertex*/
/* Reads texture co-ordinates
 */
static void TextureVertex(struct asefile *File,object_t *NewObject) {
  int i;

  if(File->NextToken!=INTEGER)
    CN_Error("integer expected");
  NewObject -> NumTVertex = File->reader.TokenValue.i;
  GetNextToken(File);

  if(NewObject->NumTVertex==0)
    return;

  if(File->NextToken!=MESH_TVERTLIST)
    CN_Error("MESH_TVERTLIST expected\n");

  GetNextToken(File);
  GetNextToken(File); /* BRACE_OPEN */

  if((NewObject -> TVerts = malloc(sizeof(vec2_t) * NewObject -> NumTVertex)) == NULL)
    CN_Error("Memory allocation failed\n");
  for(i = 0; i < NewObject -> NumTVertex; i++)
    MeshTVert(File, i, NewObject);
  GetNextToken(File); /* BRACE_CLOSE */

  /* Read Mesh Num texture face */
  if(File->NextToken!=MESH_NUMTVFACES)
    CN_Error("mesh number of texture face expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh number of texture face expected");
  GetNextToken(File);
  
  GetNextToken(File); /* MESH_TFACE_LIST */
  GetNextToken(File); /* BRACE_OPEN */

  for(i = 0; i < NewObject -> NumFaces; i++)
    MeshTFaceList(File, i, NewObject);
  GetNextToken(File); /* BRACE_CLOSE */

}

/*ColourVertex*/
/* Reads vertex colours
 */
static void ColourVertex(struct asefile *File,object_t *NewObject) {

  if(File->NextToken!=INTEGER)
    CN_Error("integer expected");
  
  if(File->reader.TokenValue.i!=0)
    CN_Warning("Handling of coloured vertices is not possible (yet)");

  GetNextToken(File);

}


/*MeshVertexList*/
/* Reads the vertex list
 */
static void MeshVertexList(struct asefile *File, int VertexID, object_t *NewObject)
{
  if(File->NextToken!=MESH_VERTEX)
    CN_Error("mesh vertex expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh vertex number expected");
 
  if(File->reader.TokenValue.i != VertexID)
    CN_Error("Inconsistent vertex list id's");
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float expected");
  NewObject ->Verts[VertexID][0] = File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float expected");
  NewObject ->Verts[VertexID][2] = File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float expected");
  NewObject ->Verts[VertexID][1] = File->reader.TokenValue.f;
  GetNextToken(File);

}

/*MeshFaceList*/
/* Reads the face list 
 */
static void MeshFaceList(struct asefile *File, int FaceID,object_t *NewObject)
{
  if(File->NextToken!=MESH_FACE)
    CN_Error("Mesh Vertex expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");
  if(File->reader.TokenValue.i != FaceID)
    CN_Error("Inconsistent vertex list id's");
  GetNextToken(File);

  if(File->NextToken!=COLON)
    CN_Error("colon face number expected");
  GetNextToken(File);

  /* Reads Point A */
  if(File->NextToken!=MESH_FACE_A)
    CN_Error("A expected");
  GetNextToken(File);

  if(File->NextToken!=COLON)
    CN_Error("colon face number expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");
  NewObject ->Faces[FaceID].VertexIndex[0] = File->reader.TokenValue.i;
  GetNextToken(File);

  /* Reads Point B */
  if(File->NextToken!=MESH_FACE_B)
    CN_Error("B expected");
  GetNextToken(File);

  if(File->NextToken!=COLON)
    CN_Error("colon face number expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");
  NewObject ->Faces[FaceID].VertexIndex[1] = File->reader.TokenValue.i;
  GetNextToken(File);

  /* Reads Point C */
  if(File->NextToken!=MESH_FACE_C)
    CN_Error("C expected");
  GetNextToken(File);

  if(File->NextToken!=COLON)
    CN_Error("colon face number expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");
  NewObject ->Faces[FaceID].VertexIndex[2] = File->reader.TokenValue.i;
  GetNextToken(File);

  /* Remainder of data in these lines does not seem to be required */
  while(File->NextToken!=MESH_MTLID)
    GetNextToken(File);
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("Integer Material ID expected");
  GetNextToken(File);
}

/*MeshTVert*/
/* Reads the face list 
 */
static void MeshTVert(struct asefile *File, int TVertID,object_t *NewObject)
{
  if(File->NextToken!=MESH_TVERT)
    CN_Error("Mesh Texture Vertex expected");
  GetNextToken(File);
  if(File->NextToken!=INTEGER)
    CN_Error("mesh face number expected");

  if(File->reader.TokenValue.i != TVertID)
    CN_Error("Inconsistent vertex list id's");
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float number expected");
  NewObject->TVerts[TVertID][0]=File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float number expected");
  NewObject->TVerts[TVertID][1]=File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float number expected");
  GetNextToken(File);

  /* discard this last value as not used, and array is only vec2_t */
}

/*MeshFaceList*/
/* Reads the face list 
 */
static void MeshTFaceList(struct asefile *File, int TFaceID,object_t *NewObject)
{
  if(File->NextToken!=MESH_TFACE)
    CN_Error("Mesh Texture face expected");
  GetNextToken(File);
 
  if(File->NextToken!=INTEGER)
    CN_Error("mesh Texture face number expected");
  if(File->reader.TokenValue.i != TFaceID)
    CN_Error("Inconsistent texture face id's");
  GetNextToken(File);

  /* Reads Point  */
  if(File->NextToken!=INTEGER)
    CN_Error("Integer expected");
  NewObject->Faces[TFaceID].TCoordIndex[0] = File->reader.TokenValue.i;
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("Integer expected");
  NewObject->Faces[TFaceID].TCoordIndex[1] = File->reader.TokenValue.i;
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("Integer expected");
  NewObject->Faces[TFaceID].TCoordIndex[2] = File->reader.TokenValue.i;
  GetNextToken(File);
}

/*MeshNormals*/
/* Reads face normals */
static void MeshNormals(struct asefile *File, int NormalID,object_t *NewObject)
{
  float         m;
  int           VertexID;
  int           i;

  if(File->NextToken!=MESH_FACENORMAL)
    CN_Error("Mesh face normal expected");
  GetNextToken(File);

  if(File->NextToken!=INTEGER)
    CN_Error("mesh Texture face number expected");
  if(File->reader.TokenValue.i != NormalID)
    CN_Error("Inconsistent normal face id's");
  GetNextToken(File);

  /* Read face normals */
  /* note these are read out of sequence, 0, 2, 1, because y and z are swapped apparently */
  if(File->NextToken!=FLOAT)
    CN_Error("float for face normal expected");
  NewObject -> Faces[NormalID].FaceNormal[0] = File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("float for face normal expected");
  NewObject -> Faces[NormalID].FaceNormal[2] =File->reader.TokenValue.f; 
  GetNextToken(File);
 
  if(File->NextToken!=FLOAT)
    CN_Error("float for face normal expected");
  NewObject -> Faces[NormalID].FaceNormal[1] =File->reader.TokenValue.f;
  GetNextToken(File);

  /*nx = ( ( y1 - y2 ) * ( z1 - z3 ) ) - ( ( z1 - z2 ) * ( y1 - y3 ));*/
  NewObject->Faces[NormalID].FaceNormal[0]=
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][1]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][1])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][2]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][2]))
      -
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][2]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][2])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][1]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][1]));

    /*ny = ( ( z1 - z2 ) * ( x1 - x3 ) ) - ( ( x1 - x2 ) * ( z1 - z3 ));*/
    NewObject->Faces[NormalID].FaceNormal[1]=
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][2]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][2])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][0]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][0]))
      -
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][0]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][0])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][2]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][2]));

    /*nz = ( ( x1 - x2 ) * ( y1 - y3 ) ) - ( ( y1 - y2 ) * ( x1 - x3 ));*/
    NewObject->Faces[NormalID].FaceNormal[2]=
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][0]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][0])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][1]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][1]))
      -
     ((NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][1]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[1]][1])
        *
      (NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[0]][0]-
       NewObject->Verts[NewObject->Faces[NormalID].VertexIndex[2]][0]));

    /* Normalise */
    m=(float) -VEC_Magnitude(NewObject->Faces[NormalID].FaceNormal);
    NewObject->Faces[NormalID].FaceNormal[0]/=m;
    NewObject->Faces[NormalID].FaceNormal[1]/=m;
    NewObject->Faces[NormalID].FaceNormal[2]/=m;

  /* Some of these values are duplicated when reading from the file */
  /* what the hell I may remove duplicates at a later data */

  /* Always 3 vertexes per face, i.e., triangle */
  for(i = 0; i < 3; i++)
  {
    if(File->NextToken!=MESH_VERTEXNORMAL)
      CN_Error("Mesh vertex normal expected");
    GetNextToken(File);
 
    if(File->NextToken!=INTEGER)
      CN_Error("vertex ID expected");
    VertexID = File->reader.TokenValue.i;
    GetNextToken(File);

    /* Note these are read as 0, 2, 1, because y and Z axis are swapped in 3D Studio Max */
    if(File->NextToken!=FLOAT)
      CN_Error("float for vertex normal expected");
    NewObject ->VertexNorms[VertexID][0] = File->reader.TokenValue.f;
    GetNextToken(File);

    if(File->NextToken!=FLOAT)
      CN_Error("float for vertex normal expected");
    NewObject ->VertexNorms[VertexID][2] = File->reader.TokenValue.f;  
    GetNextToken(File);

    if(File->NextToken!=FLOAT)
      CN_Error("float for vertex normal expected");
    NewObject ->VertexNorms[VertexID][1] = File->reader.TokenValue.f;
    GetNextToken(File);
  }
}

/*CameraObject*/
/* Routine to parse targeted cameras
 *  from ASE files.  12/2/00, Mcm  
 */
static camera_t *CameraObject(struct asefile *File) {
  camera_t *RC;

  if((RC=malloc(sizeof(camera_t)))==NULL)
    CN_Error("malloc failed");

  Expect(BRACE_OPEN);
  Expect(NODE_NAME);

  if(File->NextToken!=STRING)
    CN_Error("STRING expected");
  RC->Name=strdup(File->reader.TokenValue.s);
  GetNextToken(File);

  Expect(CAMERATYPE);
  if(File->NextToken!=TARGET)
    CN_Error("Only target cameras are supported by this parser");
  GetNextToken(File);

  NodeTM(File,&RC->P); 
  NodeTM(File,&RC->Target); 

  /* Now skip until the closing brace */
  while(File->NextToken!=BRACE_CLOSE) 
    GetNextToken(File);
  GetNextToken(File);

  Expect(BRACE_CLOSE);

  return RC;
}

/*TrackObject*/
/* Routine to parse line paths
 * from ASE files. 20/2/00 TM
 */
static shape_t *TrackObject(struct asefile *File)
{
  shape_t *t;
  int i;
  int linecount;

  if((t = malloc(sizeof(shape_t))) == NULL)
    CN_Error("Malloc failed for new track");

  Expect(BRACE_OPEN);
  Expect(NODE_NAME);

  if(File->NextToken!=STRING)
    CN_Error("STRING expected");
  t->Name = strdup(File->reader.TokenValue.s);
  GetNextToken(File);

  /* Skip the NODE TM section as this is not required */
  while(File->NextToken!= SHAPE_LINECOUNT)
    GetNextToken(File);

  /* This always seems to be 1, hence 1 line per object, made up of any number of points */
  GetNextToken(File);

  if((linecount = File->reader.TokenValue.i) != 1)
    CN_Error("Shapeobjects with more than 1 line or not correctly supported");

  GetNextToken(File);
  Expect(SHAPE_LINE);
  Expect(INTEGER);
  if(File->reader.TokenValue.i != 0)
    CN_Error("Inconsistent id's for shape line count, expected 0");
  Expect(BRACE_OPEN);

  /* I.e. a connection between last node and first node of path */
  if(File->NextToken == SHAPE_CLOSED)
  {
    t->Closed = true;
    GetNextToken(File);
  }
  else
    t->Closed = false;

  Expect(SHAPE_VERTEXCOUNT);
  t->Num = File->reader.TokenValue.i;

  /* Now allocate memory for the vertexes */
  if((t->Verts = malloc(sizeof(vec3_t) * t->Num)) == NULL)
    CN_Error("Malloc failed for vertex array in shape object");

  /* Now read in each vertex */
  for(i = 0; i < t->Num; i++)
  {
    /* Next token may be either an actuall point (knot), of original line, or an interpolated value */
    GetNextToken(File);
    if(File->NextToken != SHAPE_VERTEX_KNOT && File->NextToken != SHAPE_VERTEX_INTERP)
      CN_Error("Expected VERTEX_KNOT or VERTEX_INTERP");
    
    GetNextToken(File);
    if(File->reader.TokenValue.i != i)
      CN_Error("Inconsistent vertex id's");

    /* Get x */
    GetNextToken(File);
    t->Verts[i][0] = File->reader.TokenValue.f;

    /* Get y */
    GetNextToken(File);
    t->Verts[i][2] = File->reader.TokenValue.f;

    /* Get z */
    GetNextToken(File);
    t->Verts[i][1] = File->reader.TokenValue.f;
  }

  GetNextToken(File);
  Expect(BRACE_CLOSE);
  Expect(BRACE_CLOSE);

  return t;
}

/*GroupObject*/
/* Reads the group object details then recursively
 * reads the objects contained within it
 */
static asegroup_t *GroupObject(struct asefile *File)
{
  asegroup_t *Group;
  aseobject_t obj;
  static char charbuf[256];
  static char indent[64];
  static int t = 0;

  /* Indenting to test correct group structure */
  if(t == 0) { strcpy(indent, ""); t = 1;}

  /* Allocate memory for the group */
  if((Group = malloc(sizeof(asegroup_t))) == NULL)
    CN_Error("Memory allocation failed for new group");

  if(File->NextToken != STRING)
    CN_Error("STRING Expected");

  Group->GroupName = strdup(File->reader.TokenValue.s);
  
  GetNextToken(File);

  printf("%s-Group: %s\n", indent, Group->GroupName);
  strcat(indent, "|");

  GetNextToken(File);

  /* Start with an initial number of 10 objects within the region */
  Group->numElements = 0;
  if((Group->Objects = malloc(sizeof(aseobject_t) * 10)) == NULL)
    CN_Error("Memory allocation for ASE Group object failed");

  /* Reads objects until an invalid object type is returned, this should be a closing brace */
  while(File->NextToken != BRACE_CLOSE)
  { 
    obj = ASE_GetObject(File);
    Group->Objects[Group->numElements] = obj;
    Group->numElements++;
    /* Check if array size needs increasing */
    if(Group->numElements % GROUP_CHUNKSIZE == 0)
    {
      Group->Objects = realloc(Group->Objects, sizeof(aseobject_t) * (Group->numElements + GROUP_CHUNKSIZE));
      if(Group->Objects == NULL)
        CN_Error("Realloc failed for ase group object");
    }
  }
  
  GetNextToken(File);

  /* Now collapse the array to correct size */
  Group->Objects = realloc(Group->Objects, sizeof(aseobject_t) * Group->numElements);

  printf(charbuf, "%s-%d objects", indent, Group->numElements);

  indent[strlen(indent) - 1] = '\0';
  
  return Group;
}


/*NodeTM*/
/* Parses the NodeTM part and returns the
 *  TM_POS values.
 */
static void NodeTM(struct asefile *File,vec3_t *Result) {

  Expect(NODE_TM);

  /* Lets skip until we get to the TM_POS part */
  while(File->NextToken!=TM_POS) 
    GetNextToken(File);
  GetNextToken(File);

  /* Now start reading in the values */
  if(File->NextToken!=FLOAT)
    CN_Error("Unexpected token");
  (*Result)[0]=File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("Unexpected token");
  (*Result)[2]=File->reader.TokenValue.f;
  GetNextToken(File);

  if(File->NextToken!=FLOAT)
    CN_Error("Unexpected token");
  (*Result)[1]=File->reader.TokenValue.f;
  GetNextToken(File);

  /* Now skip until the closing brace */
  while(File->NextToken!=BRACE_CLOSE) 
    GetNextToken(File);
  GetNextToken(File);

}

/*Depend*/
/* writes text to the dependency file
 */
static void Depend(struct asefile *File, char *message)
{
  if(File->Depend != NULL)
    fprintf(File->Depend, message);
}


/*Error*/
/* A wrapper for CN_Error, but also prints lexing info. 
 */
static void Error(const char *Message,const char *File, int Line, struct asefile *asefile)
{
  char charbuf[256];
  fprintf(stderr, "-Parse Error: %s\n Line %ld\n",   asefile->FileName, asefile->reader.LineCount);
  sprintf(charbuf,"-Parse Error: %s\n Line %ld\n%s", asefile->FileName, asefile->reader.LineCount, Message);
  CN_RealError(charbuf, File, Line);
}




