/* ASE_Lexical.c
 * $Format: " * Version $ProjectVersion$"$
 * Version 0.27
 * Thomas Morrissey, tm7733@bris.ac.uk
 * 15/11/99
 * Routines read tokens from a data stream
 * 
 */ 

#ifdef WIN32
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "Lexical.h"
#include "Object.h"
#include "Materials.h"
#include "Constants.h"

/* Token identifiers */
#define STRING_START    '\"'
#define WS_NEWLINE      '\n'
#define WS_TAB          '\t'
#define WS_SPACE        ' '
#define FLOAT_TYPE      'E'
#define FLOAT_EXPN      '-'
#define FLOAT_DCML      '.'

/* Internal function prototypes */
Token_t LEX_SymbolString  (lexreader_t *reader);
void    LEX_UnRead        (lexreader_t *reader, int c);
int     LEX_Read          (lexreader_t *reader);

/*LEX_SymbolString*/
/* Reads and returns a string token */
Token_t LEX_SymbolString(lexreader_t *reader)
{
  int c;
  int i = 0;
  char SymbolString[LEX_MAX_LENGTH];

  while(1)                           /* Loop to entire length of string       */
  {
    c = LEX_Read(reader);

    if(c == WS_NEWLINE || c == EOF)  /* Newline and EOF not allowed in string */
      return BAD;
    if(i >= LEX_MAX_LENGTH)          /* String is longer than allowed limit   */
      return BAD;
    if(c == STRING_START)
      break;
    else
      SymbolString[i++] = c;         /* otherwise add character to string     */
  }
  SymbolString[i] = '\0';
  strcpy(reader->TokenValue.s, SymbolString);
  return STRING;
}


/*LEX_Read*/
/* Reads a character from the reader object depending on the stream type
 */
int LEX_Read(lexreader_t *reader)
{
  if(reader->DataType == IO)
    return getc(reader->data.io.Stream);
  else
  { 
    if(reader->data.mem.p == reader->data.mem.length || reader->data.mem.data[reader->data.mem.p] == '\0')
      return EOF;
    else
      return (int) reader->data.mem.data[(reader->data.mem.p)++];
  }
}

/*LEX_Unread*/
/* Unreads a character from the reader object depending on the stream type
 */
void LEX_UnRead(lexreader_t *reader, int c)
{
  if(reader->DataType == IO)
    ungetc(c, reader->data.io.Stream);
}

/*LEX_Initreader_mem*/
/* Initialises the lexical reader with data from an array
 */
boolean LEX_InitReaderMEM(lexreader_t *reader, char *data, long length)
{
  reader->LineCount       = 0;
  reader->data.mem.p      = 0;
  reader->data.mem.data   = data;
  reader->data.mem.length = length;
  reader->DataType        = MEM;
  return true;
}

/*LEX_Initreader_data*/
/* Initialises the lexical reader data from file io
 */
boolean LEX_InitReaderIO(lexreader_t *reader, FILE *Stream)
{
  if(Stream == NULL)
  {
    CN_Warning("Attempt to initialise lexical reader with null stream");
    return false;
  }
  
  reader->LineCount      = 0;
  reader->data.io.Stream = Stream;
  reader->DataType       = IO;
  return true;
}


/*ASE_ReadToken*/
/* A DFA (Deterministic Finite Automatia) machine */
/* Lexical Analyser to reads the next token from the buffered stream */
Token_t LEX_ReadToken(lexreader_t *reader)
{
  int State;                            /* State for lexical analysis */
  int c = 0;                            /* Stores the current character from buffered input */
  int i;                                /* Counter for string */
  char Symbol[LEX_MAX_LENGTH];          /* Will hold symbol as string */

  State = 1;                            /* Initial state is always 1 */
  i = 0;

  while(1)                              /* read until buffer is empty */
  {

    c = LEX_Read(reader);

    if(c =='\n')                        /* Line count position */
    {               
      reader->LineCount++;
      continue;
    }

    switch(State)                       /* Now analyse the state */
    {
      case 1:                    /* Start state */
      {
        if(isdigit(c))           {State = 2; Symbol[i++] = c; break;}
        else if(isspace(c))      {State = 1;                  break;}
        else if(isalpha(c))      {State = 8; Symbol[i++] = c; break;}
        else if(c == '*')        {State = 8; Symbol[i++] = c; break;}
        else if(c == '-')        {State = 2; Symbol[i++] = c; break;}
        else if(c == '+')        {State = 2; Symbol[i++] = c; break;}
        else switch(c)
        {
          case STRING_START:     {return LEX_SymbolString(reader);}
          case ':':              {return COLON;}
          case '{':              {return BRACE_OPEN;}
          case '}':              {return BRACE_CLOSE;}
          case EOF:              {return EOF;};
        }
        return BAD;
      }
      case 2:                    /* Integer  (accept state) */
      {
        if(isdigit(c))           {State = 2; Symbol[i++] = c; break;}
        else if(c == FLOAT_DCML) {State = 3; Symbol[i++] = c; break;}
        else if(c == FLOAT_TYPE) {State = 5; Symbol[i++] = c; break;}
        else                     {LEX_UnRead(reader, c);
                                  Symbol[i] = '\0';
                                  reader->TokenValue.i =  (int) atoi(Symbol);
                                  return INTEGER;}
      }
      case 3:                    /* Must be a digit, or else bad */ 
      {
        if(isdigit(c))           {State = 4; Symbol[i++] = c; break;}
        else                     {return BAD;}

      }
      case 4:                    /* Integer or real (accept state) */
      {
        if(isdigit(c))           {State = 4; Symbol[i++] = c; break;}
        else if(c == FLOAT_TYPE) {State = 5; Symbol[i++] = c; break;}
        else                     {LEX_UnRead(reader, c);
                                  Symbol[i] = '\0';
                                  reader->TokenValue.f  = (float) atof(Symbol);
                                  return FLOAT;}
      }
      case 5:                    /* Must be float, so either additional digits, or exponent */
      {
        if(isdigit(c))           {State = 7; Symbol[i++] = c; break;}
        if(c == '-' || c == '+') {State = 6; Symbol[i++] = c; break;}
        return BAD;
      }
      case 6:                    /* Must be a float, so must have minimum 1 digit for exponent */
      {
        if(isdigit(c))           {State = 7; Symbol[i++] = c; break;}
        else                     {return BAD;}
      }
      case 7:                     /* Now final digits of exponent (accept state */
      {
        if(isdigit(c))           {State = 7; Symbol[i++] = c; break;}
        else                     {LEX_UnRead(reader, c);
                                  Symbol[i] = '\0';
                                  reader->TokenValue.f = (float) atof(Symbol);
                                  return FLOAT;}
      }
      case 8:                    /* Must be an identifier */
      {
        if(isalnum(c) || c == '_')
          if(i < ID_MAX_LENGTH) {State = 8; Symbol[i++] = c; break;}
          else                   {return BAD;}
        else
                                 {LEX_UnRead(reader, c);
                                  Symbol[i] = '\0';
                                  strcpy(reader->TokenValue.s, Symbol);
                                  return IDENTIFIER;}

  	  }
    }
  }
}




