/* Materials.c
 * $Format: " * Version $ProjectVersion$"$
 * Version 0.27
 * Michael McTernan, mm7323@bris.ac.uk
 * 14/11/99
 * A store for different material types.
 *  Mainly used to store textures.
 */ 

#ifdef WIN32
  #include <windows.h>
#endif

#include <GL/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include "Constants.h"
#include "Materials.h"
#include "TGALoader.h"

material_t *MatArray=NULL;
static int         MatMax=0;
static int         MatLast=0;
static mid_t       LastApplied=-1;

/*MAT_Store*/
/* Store the new Material in the array and return 
 * the mid_t
 */
mid_t MAT_Store(material_t Material) {

  /* Create the array */
  if(MatArray==NULL || MatLast+1>=MatMax) {
    MatMax+=10;
    MatArray=realloc(MatArray,MatMax*sizeof(material_t));
    if(MatArray==NULL)
      CN_Error("realloc failed");
  }

  /* Check if there is a texture file to load */
  if(Material.Filename!=NULL) {
    glGenTextures(1,&Material.GLTextureId);
    if(!TGA_LoadTexture(Material.Path,Material.Filename,Material.GLTextureId)) {
      free(Material.Filename);
      Material.Filename=NULL;
      Material.GLTextureId=-1;
    }
  } else
    Material.GLTextureId=-1;

  /* Put the new element into the array */
  MatArray[MatLast++]=Material;

  return MatLast-1;
}

/*MAT_ApplyMaterial*/
/* Call glMaterialfv with the parameters 
 * needed for the texture. 
 */
void MAT_ApplyMaterial(mid_t MId,boolean Tint) {
  static boolean DepthMask=true;

  /* Check that there is a material*/
  if(MId==-1)
    return;

  /* Check we are not re-applying a material already */
  if(!Tint) {

    if(MId==LastApplied)
      return;

    LastApplied=MId;
  }

  /* Enable texturing if needed */
  if(MatArray[MId].Filename!=NULL && 
     MatArray[MId].GLTextureId>0) 
  {
    GLfloat DefaultDiffuse[] =  {(float)0.8,(float)0.8,(float)0.8};
   
    if(!Tint)
      glColor4f(DefaultDiffuse[0],
                DefaultDiffuse[1],
                DefaultDiffuse[2],
                MatArray[MId].Transparency);

    glBindTexture(GL_TEXTURE_2D,MatArray[MId].GLTextureId);
    glEnable(GL_TEXTURE_2D);
  } else {

    glDisable(GL_TEXTURE_2D);

    /* Set the properties */
    if(!Tint)
      glColor4f(MatArray[MId].Diffuse[0],
                MatArray[MId].Diffuse[1],
                MatArray[MId].Diffuse[2],
                MatArray[MId].Transparency);
  }
  
  if(!Tint) {
    if(MatArray[MId].Transparency<1.0) {
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      DepthMask=false;
    } else {
      glDisable(GL_BLEND);
      glDepthMask(GL_TRUE);
      DepthMask=true;
    }
  } else {
    if(!DepthMask)
      glDepthMask(GL_TRUE);
  }

}

/*MAT_InvalidateCache*/
/* Invalidate the memory of the last 
 * texture applied
 */
void MAT_InvalidateCache() {
  LastApplied=-2;  
}

/*MAT_ResidenceCheck*/
/* Check which of the textures are cached
 */
void MAT_ResidenceCheck() {
  GLuint    *Names;
  GLboolean *Status;
  int t,u;

  Names =malloc(sizeof(GLuint)*MatLast);
  Status=malloc(sizeof(GLboolean)*MatLast);

  if(!Names || !Status) 
    CN_Error("malloc failed");

  for(u=t=0;t<MatLast;t++) {
    if(MatArray[t].Filename!=NULL)
      Names[u++]=MatArray[t].GLTextureId;
  }

  glAreTexturesResident(u,Names,Status);

  for(u=t=0;t<MatLast;t++) {

    if(MatArray[t].Filename!=NULL) {
      char Buf[80];
      
      printf(Buf,"%s\t %s\n",
                   MatArray[t].Filename,
                   Status[u]?"resident":"not resident");
      u++;
    }
    
  }


  free(Names);
  free(Status);  

}


