/* Object.c
 * $Format: " * Version $ProjectVersion$"$
 * Version 0.27
 * Michael McTernan, mm7323@Bris.ac.uk
 * 14/11/99
 * Routines to cut objects against a plane
 *  which is parallel to either the x-y plane
 *  or the y-z plane.
 * NOTE: This needs a big old tidy..
 */ 

#ifdef WIN32
  #include <windows.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glut.h>
#include <stdio.h>
#include "Object.h"
#include "Materials.h"
#include "Constants.h"

//#define DEBUGGING_OBJECT_NAMES

#define REALLOC_CHUNK_SIZE  100
#define OPT_Lighting 1

/* Static functions */
static boolean Intersect(vec3_t *Result, const vec3_t P1,const vec3_t P2,const vec3_t A,const vec3_t N);


/*OBJ_CutXLT*/
/* Return a new object that is the part of
 * Object less than x.  Original Object is
 * all parts that are greater than or equal
 * to x.
 * Return NULL if there is no part of the
 * object less than x.
 * If none of Object>=x, *Object=NULL;
 */
object_t *OBJ_CutXLT(object_t **SObject,float x) {
  int face,t;
  object_t *LTX,*GEX,*Object;
#ifdef DEBUGGING_OBJECT_NAMES
  char NameBuf[4096];
#endif

  /* Make the return objects */
  LTX=calloc(sizeof(object_t),1);
  GEX=calloc(sizeof(object_t),1);
  Object=*SObject;

#ifdef DEBUGGING_OBJECT_NAMES
  sprintf(NameBuf," (%s < x=%f) ",Object->Name,x);
  LTX->Name=strdup(NameBuf);
  sprintf(NameBuf," (%s >= x=%f) ",Object->Name,x);
  GEX->Name=strdup(NameBuf);
#else
  LTX->Name=strdup(Object->Name);
  GEX->Name=strdup(Object->Name);
#endif

  /* Loop for each face (which are triangles) */
  for(face=0;face<Object->NumFaces;face++) {
    int triLTX=0;
        
    /* Loop for each point of triangle */
    for(t=0;t<3;t++) {
      if(Object->Verts[Object->Faces[face].VertexIndex[t]][0]<x)
        triLTX++;
    }

    if(triLTX>0) {
      OBJ_AddFace(LTX,
                  Object->Verts[Object->Faces[face].VertexIndex[0]],
                  Object->Verts[Object->Faces[face].VertexIndex[1]],
                  Object->Verts[Object->Faces[face].VertexIndex[2]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[0]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[1]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[2]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[0]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[1]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[2]],
                  Object->NumTVertex>0,
                  Object->Faces[face].FaceNormal
                 );
    } 

    if(triLTX<3) {
      OBJ_AddFace(GEX,
                  Object->Verts[Object->Faces[face].VertexIndex[0]],
                  Object->Verts[Object->Faces[face].VertexIndex[1]],
                  Object->Verts[Object->Faces[face].VertexIndex[2]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[0]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[1]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[2]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[0]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[1]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[2]],
                  Object->NumTVertex>0,
                  Object->Faces[face].FaceNormal
                 );
    }

  }

  /* Check for cases where no splitting was done */
  if(LTX->NumFaces==0) {
    OBJ_Free(LTX);
    OBJ_Free(GEX);
    LTX=NULL;
    GEX=NULL;
    return NULL;
  }

  if(GEX->NumFaces==0) {
    OBJ_Free(LTX);
    OBJ_Free(GEX);
    LTX=NULL;
    GEX=NULL;
    *SObject=NULL;
    return Object;
  }

  /* We need to reduce the arrays to the correct size */
  GEX->Faces=realloc(GEX->Faces,sizeof(triangle_t)*GEX->NumFaces);
  GEX->Verts=realloc(GEX->Verts,sizeof(vec3_t)*GEX->NumVertex);
  GEX->TVerts=realloc(GEX->TVerts,sizeof(vec2_t)*GEX->NumTVertex);
  GEX->VertexNorms=realloc(GEX->VertexNorms,sizeof(vec3_t)*GEX->NumVertex);
  LTX->Faces=realloc(LTX->Faces,sizeof(triangle_t)*LTX->NumFaces);
  LTX->Verts=realloc(LTX->Verts,sizeof(vec3_t)*LTX->NumVertex);
  LTX->TVerts=realloc(LTX->TVerts,sizeof(vec2_t)*LTX->NumTVertex);
  LTX->VertexNorms=realloc(LTX->VertexNorms,sizeof(vec3_t)*LTX->NumVertex);
 
  /* Assign over the material ID */
  GEX->MId=Object->MId;
  LTX->MId=Object->MId;

  /* Splitting was done */
  OBJ_Free(Object);
  *SObject=GEX;
  return LTX;
  
}


/*OBJ_CutZLT*/
/* Return a new object that is the part of
 * Object less than z.  Original Object is
 * all parts that are greater than or equal
 * to z.
 * Return NULL if there is no part of the
 * object less than z.
 * If none of Object>=z, *Object=NULL;
 */
object_t *OBJ_CutZLT(object_t **SObject,float z) {
  int face,t;
  object_t *LT,*GE,*Object;
#ifdef DEBUGGING_OBJECT_NAMES
  char NameBuf[4096];
#endif

  /* Make the return objects */
  LT=calloc(sizeof(object_t),1);
  GE=calloc(sizeof(object_t),1);
  Object=*SObject;

  /* Some things always get copied accross */
  LT->MId=Object->MId;
  GE->MId=Object->MId;

#ifdef DEBUGGING_OBJECT_NAMES
  sprintf(NameBuf," (%s < z=%f) ",Object->Name,z);
  LT->Name=strdup(NameBuf);
  sprintf(NameBuf," (%s >= z=%f) ",Object->Name,z);
  GE->Name=strdup(NameBuf);
#else
  LT->Name=strdup(Object->Name);
  GE->Name=strdup(Object->Name);
#endif

  /* Loop for each face (which are triangles) */
  for(face=0;face<Object->NumFaces;face++) {
    int triLT=0;
    
    /* Loop for each point of triangle */
    for(t=0;t<3;t++) {
      if(Object->Verts[Object->Faces[face].VertexIndex[t]][2]<z)
        triLT++;
    }

    if(triLT>0) {
      OBJ_AddFace(LT,
                  Object->Verts[Object->Faces[face].VertexIndex[0]],
                  Object->Verts[Object->Faces[face].VertexIndex[1]],
                  Object->Verts[Object->Faces[face].VertexIndex[2]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[0]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[1]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[2]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[0]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[1]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[2]],
                  Object->NumTVertex>0,
                  Object->Faces[face].FaceNormal
                 );
    }

    if(triLT<3) {
      OBJ_AddFace(GE,
                  Object->Verts[Object->Faces[face].VertexIndex[0]],
                  Object->Verts[Object->Faces[face].VertexIndex[1]],
                  Object->Verts[Object->Faces[face].VertexIndex[2]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[0]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[1]],
                  Object->VertexNorms[Object->Faces[face].VertexIndex[2]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[0]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[1]],
                  Object->TVerts[Object->Faces[face].TCoordIndex[2]],
                  Object->NumTVertex>0,
                  Object->Faces[face].FaceNormal
                 );
    }
  }

  /* Check for cases where no splitting was done */
  if(LT->NumFaces==0) {
    OBJ_Free(LT);
    OBJ_Free(GE);
    return NULL;
  }

  if(GE->NumFaces==0) {
    OBJ_Free(LT);
    OBJ_Free(GE);
    *SObject=NULL;
    return Object;
  }

  /* We need to reduce the arrays to the correct size */
  GE->Faces=realloc(GE->Faces,sizeof(triangle_t)*GE->NumFaces);
  GE->Verts=realloc(GE->Verts,sizeof(vec3_t)*GE->NumVertex);
  GE->TVerts=realloc(GE->TVerts,sizeof(vec2_t)*GE->NumTVertex);
  GE->VertexNorms=realloc(GE->VertexNorms,sizeof(vec3_t)*GE->NumVertex);
  LT->Faces=realloc(LT->Faces,sizeof(triangle_t)*LT->NumFaces);
  LT->Verts=realloc(LT->Verts,sizeof(vec3_t)*LT->NumVertex);
  LT->TVerts=realloc(LT->TVerts,sizeof(vec2_t)*LT->NumTVertex);
  LT->VertexNorms=realloc(LT->VertexNorms,sizeof(vec3_t)*LT->NumVertex);

  /* Assign over the material ID */
  GE->MId=Object->MId;
  LT->MId=Object->MId;

  /* Splitting was done */
  OBJ_Free(Object);
  *SObject=GE;
  return LT;
  
}

/*OBJ_Display*/
/* Draws an object using OpenGL calls.
 * Optimised by loop unrolling and
 *  moving conditionaly out of loops.
 */
void OBJ_Display(object_t *Object,
                 boolean   FaceLight,
                 boolean   VertexLight,
                 boolean   Tint)
{
  int i;


  MAT_ApplyMaterial(Object->MId,Tint);

  if(OPT_Lighting) {

    if(FaceLight)
      glShadeModel(GL_FLAT);
    else
      glShadeModel(GL_SMOOTH);

    /* Do the case with texture co-ordinates */
    if(Object->MId>0 && Object->NumTVertex>0) {

      glBegin(GL_TRIANGLES);

      /* Put in the faces */
      for(i=0;i<Object->NumFaces;i++) {
  
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[0]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[0]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );

        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[0]]);

        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[0]]);
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[1]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[1]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );
        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[1]]);

        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[1]]);
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[2]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[2]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );
        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[2]]);

        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[2]]);
      }
      glEnd ();
   
    } else {
      glBegin(GL_TRIANGLES);
      /* Put in the faces */
      for(i=0;i<Object->NumFaces;i++) {

        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[0]]);
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[0]]);
        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[1]]);
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[1]]);
        if(FaceLight)
          glNormal3fv(Object->Faces[i].FaceNormal);
        else 
          glNormal3fv(Object->VertexNorms[Object->Faces[i].VertexIndex[2]]);
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[2]]);
      }
      glEnd();
    }

  } else {

    /* Do the case with texture co-ordinates */
    if(Object->MId>0 && Object->NumTVertex>0) {

      MAT_ApplyMaterial(Object->MId,Tint);

      glBegin(GL_TRIANGLES);

      /* Put in the faces */
      for(i=0;i<Object->NumFaces;i++) {
  
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[0]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[0]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[0]]);
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[1]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[1]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[1]]);
        glTexCoord2f((Object->TVerts[Object->Faces[i].TCoordIndex[2]][0]*
                       MAT_Get(Object->MId).uTile)+
                       MAT_Get(Object->MId).uOffset,
                      (Object->TVerts[Object->Faces[i].TCoordIndex[2]][1]*
                       MAT_Get(Object->MId).vTile)+ 
                       MAT_Get(Object->MId).vOffset
                      );
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[2]]);
      }
      glEnd();
    } else {
      glBegin(GL_TRIANGLES);
      for(i=0;i<Object->NumFaces;i++) {
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[0]]);
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[1]]);
        glVertex3fv(Object->Verts[Object->Faces[i].VertexIndex[2]]);
      }
      glEnd ();
    }
  }

}



/*OBJ_Free*/
/* Free object and all associated memory
 */
void OBJ_Free(object_t *O) {
  free(O->Name);
  free(O->Verts);
  free(O->TVerts);
  free(O->VertexNorms);
  free(O->Faces);
  /*HACK!! What about the materials that might be freeable?? */
  free(O);
}


/*OBJ_FindCentre*/
/* Average the x,y,z co-ords for the vertices.
 */
void OBJ_FindCentre(object_t *O,vec3_t *Result) {
  int t;

  (*Result)[0]=O->Verts[0][0];
  (*Result)[1]=O->Verts[0][1];
  (*Result)[2]=O->Verts[0][2];

  for(t=1;t<O->NumVertex;t++) {
      (*Result)[0]+=O->Verts[t][0];
      (*Result)[1]+=O->Verts[t][1];
      (*Result)[2]+=O->Verts[t][2];
  }

  (*Result)[0]/=O->NumVertex;
  (*Result)[1]/=O->NumVertex;
  (*Result)[2]/=O->NumVertex;

}

/*OBJ_FindMinY*/
/* Find the lowest Y co-ordinate in an object 
 */
float OBJ_FindMinY(object_t *O) {
  float y;
  int t;

  y=O->Verts[0][1];

  for(t=1;t<O->NumVertex;t++) 
    y=_min(y,O->Verts[t][1]);

  return y;
}


/*OBJ_FindMaxY*/
/* Find the largest Y co-ordinate in an object 
 */
float OBJ_FindMaxY(object_t *O) {
  float y;
  int t;

  y=O->Verts[0][1];

  for(t=1;t<O->NumVertex;t++) 
    y=_max(y,O->Verts[t][1]);

  return y;
}


/*OBJ_AddFace*/
/* Add a face to an object
 * This function takes lots of parameters to make it 
 *  easy to write the triangle cutter...
 */
void OBJ_AddFace(object_t *O,
                 vec3_t p1,vec3_t p2,vec3_t p3,
                 vec3_t vn1,vec3_t vn2,vec3_t vn3,
                 vec2_t tc1,vec2_t tc2,vec2_t tc3,
                 boolean AddTexture,
                 vec3_t FNormal) 
{
  triangle_t Triangle;
  vec3_t Points[3],Norms[3];
  vec2_t Texture[3];
  boolean Found;
  int t,v;

  /* Make things easier to manipulate */
  for(t=0;t<3;t++) {
    Points[0][t]=p1[t];
    Points[1][t]=p2[t];
    Points[2][t]=p3[t];
    Norms[0][t]=vn1[t];
    Norms[1][t]=vn2[t];
    Norms[2][t]=vn3[t];
    if(AddTexture && t<2) {  /* vec2_t */
      Texture[0][t]=tc1[t];
      Texture[1][t]=tc2[t];
      Texture[2][t]=tc3[t];
    }
    Triangle.FaceNormal[t]=FNormal[t];
  }

  /* Now search the vertex list to see if each of the vertices are
   * already listed */
  for(v=0;v<3;v++) {

    Found=false;
    for(t=0;t<O->NumVertex && !Found;t++) {
      if(Points[v][0]==O->Verts[t][0] &&
         Points[v][1]==O->Verts[t][1] &&
         Points[v][2]==O->Verts[t][2] &&
         Norms[v][0]==O->VertexNorms[t][0] &&
         Norms[v][1]==O->VertexNorms[t][1] &&
         Norms[v][2]==O->VertexNorms[t][2]) 
      {
        Found=true;
        Triangle.VertexIndex[v]=t;
      }
    }
 
    /* If we don't have this vertex, we must add it */
    if(!Found) {
      if((O->NumVertex % REALLOC_CHUNK_SIZE)==0) {
        O->Verts=realloc(O->Verts,
                         sizeof(vec3_t)*(O->NumVertex+REALLOC_CHUNK_SIZE));
        O->VertexNorms=realloc(O->VertexNorms,
                         sizeof(vec3_t)*(O->NumVertex+REALLOC_CHUNK_SIZE));
      }
      O->NumVertex++;

      if(O->Verts==NULL || O->VertexNorms==NULL)
        CN_Error("realloc failed.");

      /* Add the vertex */
      for(t=0;t<3;t++) {
        O->Verts[O->NumVertex-1][t]=Points[v][t];
        O->VertexNorms[O->NumVertex-1][t]=Norms[v][t];
      }

      Triangle.VertexIndex[v]=O->NumVertex-1;

    } 

    /* Do the same for texture co-ordinates if required */
    if(AddTexture) {

      Found=false;
      for(t=0;t<O->NumTVertex && !Found;t++) {
        if(Texture[v][0]==O->TVerts[t][0] && 
           Texture[v][1]==O->TVerts[t][1])
        {        
          Found=true;
          Triangle.TCoordIndex[v]=t;
        }
      }
 
      /* If we don't have this texture vertex, we must add it */
      if(!Found) {
        if((O->NumTVertex % REALLOC_CHUNK_SIZE)==0) 
        {
          O->TVerts=realloc(O->TVerts,
                      sizeof(vec2_t)*(O->NumTVertex+REALLOC_CHUNK_SIZE));
        }
        
        O->NumTVertex++;

        if(O->TVerts==NULL)
          CN_Error("realloc failed.");

        /* Add the texture vertex */
        O->TVerts[O->NumTVertex-1][0]=Texture[v][0];
        O->TVerts[O->NumTVertex-1][1]=Texture[v][1];

        /* Now set the vertex */
        Triangle.TCoordIndex[v]=O->NumTVertex-1;

      }
    }
  }

  if((O->NumFaces % REALLOC_CHUNK_SIZE)==0) {
    O->Faces=realloc(O->Faces,
                     sizeof(triangle_t)*(O->NumFaces+REALLOC_CHUNK_SIZE));
  }
  O->NumFaces++;
  O->Faces[O->NumFaces-1]=Triangle;

}


/*OBJ_Intersect*/
/* Returns true if the line between A and A+N
 *  intersects any plane in the object.
 * Additionally, if there is an intersection,
 *  the intersection point is written to *Result
 */ 
boolean OBJ_Intersect(vec3_t *Result,object_t *O,const vec3_t A,const vec3_t N) {
  int t;
  vec3_t B;

  B[0]=A[0]-N[0];
  B[1]=A[1]-N[1];
  B[2]=A[2]-N[2];

  for(t=0;t<O->NumFaces;t++) {

    /* Check each face plane for intersection */
    if(Intersect(Result,A,B,
         O->Verts[O->Faces[t].VertexIndex[0]],
         O->Faces[t].FaceNormal) )
    {
      /* If intersection point is in the rectangle defined
       *   by the three model vertices, assume a collision
       */
      vec3_t box1,box2;
      boolean InBox;
      int u;

      box1[0]=box2[0]=O->Verts[O->Faces[t].VertexIndex[0]][0];      
      box1[1]=box2[1]=O->Verts[O->Faces[t].VertexIndex[0]][1];      
      box1[2]=box2[2]=O->Verts[O->Faces[t].VertexIndex[0]][2];
      
      for(u=1;u<3;u++) {
        box1[0]=_min(box1[0],O->Verts[O->Faces[t].VertexIndex[u]][0]);      
        box1[1]=_min(box1[1],O->Verts[O->Faces[t].VertexIndex[u]][1]);      
        box1[2]=_min(box1[2],O->Verts[O->Faces[t].VertexIndex[u]][2]);
        box2[0]=_max(box2[0],O->Verts[O->Faces[t].VertexIndex[u]][0]);      
        box2[1]=_max(box2[1],O->Verts[O->Faces[t].VertexIndex[u]][1]);      
        box2[2]=_max(box2[2],O->Verts[O->Faces[t].VertexIndex[u]][2]);
      }

      InBox=true;

      for(u=0;u<3 && InBox;u++) {
        if(box1[u]>(*Result)[u] || box2[u]<(*Result)[u])
          InBox=false;
      }

      if(InBox) {
        char Buf[512];
        printf(Buf,"Collision with %s\n\a",O->Name);
        return true;
      }
    }

  }
  return false;
}


/*OBJ_Duplicate*/
/* Return a new object that is a copy of O
 */
object_t *OBJ_Duplicate(object_t *O) {
  object_t *Object;
#ifdef DEBUGGING_OBJECT_NAMES
  char NameBuf[4096];
#endif

  /* Make a return object */
  if((Object=malloc(sizeof(object_t)))==NULL)
    CN_Error("malloc failed");

  /* Copy all the fields */
  memcpy(Object,O,sizeof(object_t));


#ifdef DEBUGGING_OBJECT_NAMES
  sprintf(NameBuf," (dup %s) ",O->Name);
  Object->Name=strdup(NameBuf);
#else
  Object->Name=strdup(O->Name);
#endif

  /* We need to reduce the arrays to the correct size */
  Object->Faces=malloc(sizeof(triangle_t)*O->NumFaces);
  Object->Verts=malloc(sizeof(vec3_t)*O->NumVertex);
  Object->TVerts=malloc(sizeof(vec2_t)*O->NumTVertex);
  Object->VertexNorms=malloc(sizeof(vec3_t)*O->NumVertex);
 
  /* Copy over the data */
  memcpy(Object->Faces,O->Faces,sizeof(triangle_t)*O->NumFaces);
  memcpy(Object->Verts,O->Verts,sizeof(vec3_t)*O->NumVertex);
  memcpy(Object->TVerts,O->TVerts,sizeof(vec2_t)*O->NumTVertex);
  memcpy(Object->VertexNorms,O->VertexNorms,sizeof(vec3_t)*O->NumVertex);

  return Object;
  
}


/*Intersect*/
/* Returns true if the line between P1 and P2
 *  intersects the plane.
 * Additionally, if there is an intersection,
 *  the intersection point is written to *Result
 */ 
static boolean Intersect(vec3_t *Result,
                         const vec3_t P1,
                         const vec3_t P2,
                         const vec3_t A,
                         const vec3_t N) 
{
  vec3_t a;
  float f,u;
  
  /* Calculate the vector for the line */
  (*Result)[0]=P2[0]-P1[0];
  (*Result)[1]=P2[1]-P1[1];
  (*Result)[2]=P2[2]-P1[2];

  /* Check for parallel situation */
  if((u=VEC_dotProduct(*Result,N))==0)
    return false;
 
  a[0]=A[0]-P1[0];
  a[1]=A[1]-P1[1];
  a[2]=A[2]-P1[2];

  f=VEC_dotProduct(a,N);

  (*Result)[0]*=f/u;
  (*Result)[1]*=f/u;
  (*Result)[2]*=f/u;

  (*Result)[0]+=P1[0];
  (*Result)[1]+=P1[1];
  (*Result)[2]+=P1[2];
 
  /* Now check that intersection point is between P1 & P2 */
  if((*Result)[0]>=_min(P1[0],P2[0]) &&
     (*Result)[0]<=_max(P1[0],P2[0]) &&
     (*Result)[1]>=_min(P1[1],P2[1]) &&
     (*Result)[1]<=_max(P1[1],P2[1]) &&
     (*Result)[2]>=_min(P1[2],P2[2]) &&
     (*Result)[2]<=_max(P1[2],P2[2])) {
    return true;
  }

  return false;
}


