/* TGALoader.c
 * $Format: " * Version $ProjectVersion$"$
 * Version 0.27
 * Michael McTernan, mm7323@bris.ac.uk
 * 23/11/99
 * Loads in a 24 or 32 bit TGA file and
 *  pumps it into OpenGL
 */

#ifdef WIN32
  #include <windows.h>
#endif 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "Constants.h"
#include "TGALoader.h"

static int OPT_GraphicsQuality=4;

/* Global variable */
static GLenum texFormat;

/* Static prototypes */
static int checkSize(int x);
static unsigned char *getRGBA (FILE *s, int size);
static unsigned char *getRGB (FILE *s, int size);


/*TGA_Load*/
/* Loads a 24bit or 32 bit targa file and
 * return it's dimensions and the data in a buffer
 */
unsigned char *TGA_Load(int *Width,int *Height,int *Bits, const char *Path, const char *FileName)
{
  unsigned char type[4];
  unsigned char info[7];
  unsigned char *imageData = NULL;
  char charbuf[256];
  char fn[50];
  int size;
  FILE *s;
	
  sprintf(fn, "%s%s", Path, FileName);

  /* Open the input file */
  if((s=fopen (fn, "rb"))==NULL)
  {
    sprintf(charbuf, "TGALoader: Unable to open file %s", fn);
    CN_Warning(charbuf);
    perror(NULL);
    return NULL;
  }

  /* Read in colour map and image type info */
  fread(&type,sizeof (char), 3, s); 
  fseek(s,12,SEEK_SET);		
  /* Skip rest of header */
  fread(&info,sizeof (char), 6, s);

  /* Make sure that the colourmap is 0 and type is 2, unmapped RGB */
  if (type[1] != 0 || type[2] != 2)
  {
    sprintf(charbuf,"Bad image type: %s", fn);
    CN_Error(charbuf);
  }

  *Width=info[0]+info[1]*256; 
  *Height=info[2]+info[3]*256;
  *Bits=info[4]; 

  /* Calculate the image size */
  size=(*Width)*(*Height);

  /* Read the image data */
  switch(*Bits) {
  case 32:
    imageData=getRGBA(s,size);
    break;

  case 24:
    imageData=getRGB(s,size);	
    break;

  default:
    sprintf(charbuf,"File: %s Bits: %d\n",FileName,*Bits);
    CN_Warning(charbuf);
    CN_Warning("Image has invalid number of bits");
    break;
  }
	
  fclose(s);

  /* Check that the image was loaded */
  if(imageData==NULL)
    CN_Warning("Failed to load image data");

  return imageData;
}


/*TGA_LoadTexture*/
/* Loads a 24bit or 32 bit targa file and puts it into OpenGL
 *  id is the texture ID to bind to.
 */
boolean TGA_LoadTexture(const char *Path, const char *FileName, int id)
{
  static long LoadedBytes=0;
  unsigned char *imageData = NULL;
  int imageWidth, imageHeight;
  int imageBits;
  int maxWidth,maxHeight;

  imageData=TGA_Load(&imageWidth,&imageHeight,&imageBits, Path, FileName);

  if(imageData == NULL)
    return false;

  /* Check the dimensions of the image are acceptable */
  if(!checkSize(imageWidth) || !checkSize(imageHeight)) {
    char charbuf [64];
    sprintf(charbuf ,"%s TGA image dimensions not a power of 2. Texture discarded\n",FileName);
    CN_Warning(charbuf);
    return false;
  }

  if(OPT_GraphicsQuality<2) { 
    glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxWidth);
    glGetIntegerv(GL_MAX_TEXTURE_SIZE,&maxHeight);

    if(imageWidth>maxWidth || imageHeight>maxHeight) {
      char Buf[512];
      unsigned char *newImage=
                     malloc(maxWidth*
                            maxHeight*
                            (texFormat==GL_RGB?3:4));

      if(newImage==NULL) 
        CN_Error("malloc failed");
    
        gluScaleImage(texFormat,
                    imageWidth,
                    imageHeight,
                    GL_UNSIGNED_BYTE,
                    imageData,
                    imageWidth>maxWidth?maxWidth:imageWidth,
                    imageHeight>maxHeight?maxHeight:imageHeight,
                    GL_UNSIGNED_BYTE,
                    newImage); 
      free(imageData);
      imageData=newImage;
      imageWidth=imageWidth>maxWidth?maxWidth:imageWidth,
      imageHeight=imageHeight>maxHeight?maxHeight:imageHeight,
    
      sprintf(Buf,"Warning: Texture \"%s\" is too large for texturing,"
                  " and has been scaled down to %dx%d.",
                  FileName,imageWidth,imageHeight);
      CN_Warning(Buf);
    }
  }

  /* OpenGL texture variables */
  glBindTexture(GL_TEXTURE_2D, id);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  //glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  switch(OPT_GraphicsQuality) {
  case 0:
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
    break;

  case 1:
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    break;
 
  case 2:
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
    break;

  case 4:
  case 3:
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    break; 

  }

  if(OPT_GraphicsQuality>=2) {
    gluBuild2DMipmaps(GL_TEXTURE_2D,
                      texFormat,
                      imageWidth, 
                      imageHeight,
                      texFormat,
                      GL_UNSIGNED_BYTE,
                      imageData); 

    while(imageWidth>1 && imageHeight>1) {
      LoadedBytes+= (texFormat==GL_RGB?3:4) * imageWidth * imageHeight;  
      imageWidth/=2;
      imageHeight/=2;
    }

  } else {
    glTexImage2D(GL_TEXTURE_2D,
                 0,
                 texFormat,
                 imageWidth,
                 imageHeight,
                 0,
                 texFormat,
                 GL_UNSIGNED_BYTE,
                 imageData);
    LoadedBytes+= (texFormat==GL_RGB?3:4) * imageWidth * imageHeight;  
  }

  /* Free data as it has been uploaded */
  free(imageData);

//  printf("%ld k \n",LoadedBytes/1024);

  return true;
}


/*checkSize*/
/* Make sure the image is ok without scaling
 */
static int checkSize(int x) {
  return (x==2 || x==4 || x==8 || x==16 || x==32 || x==64 || x==128 || x==256 || x==512);
}


/*getRGBA*/
/* Reads in RGBA data and returns the same, 32 bit image
 */
static unsigned char *getRGBA (FILE *s, int size) {
  unsigned char *rgba;
  unsigned char temp;
  int bread;
  int i;

  /* Allocate memory for image */
  if((rgba=malloc(size*4))==NULL)
    CN_Error("malloc failed");

  /* Read data */
  bread=fread(rgba,sizeof(unsigned char),size*4,s); 
	
  /* Make sure all data was read */
  if(bread!=size*4) 
    CN_Error("Failed to read whole image");
	
  for(i=0;i<size*4;i+=4) {
    temp = rgba[i];
    rgba[i] = rgba[i + 2];
    rgba[i + 2] = temp;
  }

  texFormat = GL_RGBA;
  return rgba;
}


/*getRGB*/
/* Reads in RGB data and returns RGB, 24bit image
 */
static unsigned char *getRGB (FILE *s, int size) {
  unsigned char *rgb;
  unsigned char temp;
  int bread;
  int i;

  /* Allocate memory for the image */
  if((rgb=malloc(size*3))==NULL)
    CN_Error("malloc failed\n"); 

  /* Read the image data */
  bread=fread(rgb, sizeof (unsigned char), size * 3, s); 

  /* Make sure all data was read */
  if(bread!=size*3)
    CN_Error("Failed to read whole image");

  for(i=0;i<size*3;i+=3) {
    temp = rgb[i];
    rgb[i] = rgb[i + 2];
    rgb[i + 2] = temp;
  }
	
  texFormat=GL_RGB;

  return rgb;
}





